// Comm.cpp: implementation of the CComm class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Comm.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CComm::CComm()
{

	hComm = INVALID_HANDLE_VALUE;

}

CComm::~CComm()
{

	if(hComm != INVALID_HANDLE_VALUE)
		CloseHandle(hComm);

}

BOOL CComm::IsOpen() 
{
	if(hComm == INVALID_HANDLE_VALUE)
		return(FALSE);
	else
		return(TRUE);
}

int CComm::OpenPort(short port) {
	// open the comm port
	COMMTIMEOUTS cto;
	DCB dcb;
	CString csTemp;


#ifdef LOGGING
	FILE *fp;
#endif

#ifdef LOGGING
	fp = fopen(SCOMM_LOG_FILE,"a");
	if(fp) fprintf(fp,"OpenPort() called with %d\n",port);
#endif

	csTemp.Format(_T("COM%d:"),port);

	if(hComm != INVALID_HANDLE_VALUE) {
#ifdef LOGGING
		if(fp) {
			fprintf(fp,"OpenPort() called but port already open\n");
		}
#endif

		CloseHandle(hComm);
		hComm = INVALID_HANDLE_VALUE;
	}

	
	FillMemory(&dcb,sizeof(dcb),0);

	hComm = CreateFile (csTemp,
		GENERIC_READ | GENERIC_WRITE,
		0,
		NULL,
		OPEN_EXISTING,
//		FILE_FLAG_OVERLAPPED,
		0,
		NULL);

	if(hComm == INVALID_HANDLE_VALUE)
	{

#ifdef LOGGING
		if(fp) fprintf(fp,"OpenPort() failed to open port %s\n", csTemp);
		fclose(fp);
#endif

		CloseHandle(hComm);
		return(1);
	}
	
	dcb.fBinary = 1;	
	cto.ReadIntervalTimeout = MAXDWORD;
	cto.ReadTotalTimeoutConstant = 500;
	cto.ReadTotalTimeoutMultiplier = MAXDWORD;
	cto.WriteTotalTimeoutConstant = 500;
	cto.WriteTotalTimeoutMultiplier = 200;

	dcb.BaudRate = 19200;
	dcb.fBinary = TRUE;
	dcb.fParity = FALSE;
	dcb.fOutxCtsFlow = FALSE;
	dcb.fOutxDsrFlow = FALSE;
	dcb.fDsrSensitivity = FALSE;
	dcb.fTXContinueOnXoff = FALSE;
	dcb.ByteSize = 8;
	dcb.Parity = NOPARITY;
	dcb.StopBits = TWOSTOPBITS;
	
	dcb.fDtrControl = DTR_CONTROL_ENABLE;
	dcb.fRtsControl = RTS_CONTROL_ENABLE; //works on usa
//	dcb.fRtsControl = RTS_CONTROL_DISABLE;
	
	dcb.DCBlength = sizeof(dcb);

	//Sets the port to the specifications in the DCB or reports an error
	if(!SetCommState(hComm,&dcb))
	{
#ifdef LOGGING
		if(fp) {
			fprintf(fp,"OpenPort() failed to SetCommState\n");
			fclose(fp);
		}
#endif
		CloseHandle(hComm);
		return(3);
	}
	else 
	{
//		PurgeComm(hComm,PURGE_TXCLEAR);	//clears any remaining chars in the output buffer
//		PurgeComm(hComm,PURGE_RXCLEAR);	//clears any remaining chars in the input buffer
		if(!SetCommTimeouts(hComm,&cto))
		{
#ifdef LOGGING
			if(fp) {
				fprintf(fp,"OpenPort() failed to set sTimeouts\n");
				fclose(fp);
			}
#endif
			CloseHandle(hComm);
			return(4);
		}
		else {
			SetCommMask(hComm, EV_RXCHAR); // sigma
#ifdef LOGGING
			if(fp) {
				fprintf(fp,"OpenPort() was successful\n");
				fclose(fp);
			}
#endif

			return(0);
		}
	}
	
	return 0;
}

int CComm::ClosePort()
{

	CloseHandle(hComm);
	return 0;
}

int CComm::ReadBytes(short sNumBytes, short sTimeout) {

#ifdef LOGGING
	FILE *fp;
#endif

	DWORD dwRead;				//identifier for Read
	DWORD dwError;
	char chRead[3];			//character read
	int x = 0;					//used for looping and is returned at the end
	int sTimeouts;
	OVERLAPPED lpOverlapped;	//overlapped structure for ReadFile
	

	//cReadBuffer[0] = 0;			//reinitialization clear any data prev
	sReadBufferLen = 0;
	
	sTimeouts = 0;				// reset sTimeout counter

	lpOverlapped.Offset = 0;
	lpOverlapped.OffsetHigh = 0;
	lpOverlapped.hEvent = NULL;

#ifdef LOGGING
	fp = fopen(SCOMM_LOG_FILE,"a");
	if(fp) {
		fprintf(fp,"ReadBytes() called with %d %d\n", sNumBytes, sTimeout);
	}
#endif

	if(hComm == INVALID_HANDLE_VALUE) {
#ifdef LOGGING
		if(fp) {
			fprintf(fp,"ReadBytes() called but comm port closed\n");
			fclose(fp);
		}
#endif
		return 3;
	}
	
	//set event mask to wait for a char to be pressed or report an error

//	if (!SetCommMask(hComm, EV_RXCHAR)) {
//		dwError = GetLastError();

	//SetCommMask(hComm, EV_RXCHAR);
	if(0) {

#ifdef LOGGING
		if(fp) {
			fprintf(fp,"ReadBytes() failed to set event mask (%d)\n",dwError);
			fclose(fp);
		}
#endif
	   return 1;
	}


	while (sReadBufferLen < sNumBytes) 
	{
		do
		{
			//read a single character at a time from the com port
			//if (ReadFile(hComm, &chRead, 1, &dwRead, &lpOverlapped))
			if (ReadFile(hComm, &chRead, 1, &dwRead, NULL))
			{
				if(dwRead)
				{
					cReadBuffer[sReadBufferLen++] = chRead[0];
					x++;
				} 
					
				if(!dwRead) {
					sTimeouts++;
					ClearCommError(hComm,&dwError,NULL);
				}

				if((sTimeouts > (sTimeout * 2)) && (sTimeout != 0)) {

#ifdef LOGGING
					if(fp) {
						fprintf(fp,"ReadBytes() sTimeout error\n");
						fclose(fp);
					}
#endif LOGGING
					//SetCommMask(hComm,0);
					return 2;
				}
			}
			else {

				dwError = GetLastError();
	
#ifdef LOGGING
				if(fp) {
					fprintf(fp,"ReadBytes() unspecified error (%d) occured\n",dwError);
					fclose(fp);
				}
#endif LOGGING
				//SetCommMask(hComm,0);
				return 5;
			}

		} while (dwRead & (sReadBufferLen < sNumBytes));
	}

	//PurgeComm(hComm,PURGE_TXCLEAR);	//clears any remaining chars in the input buffer
	//SetCommMask(hComm,0);
	cReadBuffer[sReadBufferLen] = 0;

#ifdef LOGGING
	if(fp) {
		fprintf(fp,"ReadBytes() returned with buffer contents [%s] (%d)\n",cReadBuffer,cReadBuffer[0]);
		fprintf(fp,"ReadBytes() had %d sTimeouts and got %d chars\n",sTimeouts,x);
		fclose(fp);
	}
#endif

	return 0;
}

int CComm::WriteByte(short sByte) {
#ifdef LOGGING
	FILE *fp;
	
#endif


	// write a string to the port
	DWORD write;		//number of chars to write
	DWORD errors;
	DWORD sWriteSize; 
		
	write = 1;
	sWriteSize = 1;

#ifdef LOGGING
	fp = fopen(SCOMM_LOG_FILE,"a");
	if(fp) {
		fprintf(fp,"WriteByte() called with %d %c\n", byte, byte);
	}
#endif

	if(hComm == INVALID_HANDLE_VALUE) {
#ifdef LOGGING
		if(fp) {
			fprintf(fp,"WriteBuffer() called but comm port not open\n");
			fclose(fp);
		}
#endif

		return(11);
	}
		
			if (!WriteFile(hComm,&sByte,sWriteSize,&sWriteSize,NULL))
			{
				if (GetLastError() == ERROR_IO_PENDING)
				{
					ClearCommError(hComm,&errors,NULL);
			
				}
			}
			else {

#ifdef LOGGING
				if(fp) {
					//SetCommMask(hComm,0);
					fprintf(fp,"WriteByte() failed\n");
					fclose(fp);
				}
#endif

				return 1;
			}
				
		
#ifdef LOGGING
	if(fp) {
		fprintf(fp,"WriteByte() succeeded\n");
		fclose(fp);
	}
#endif

	return 0;
	
}
