// MechThread.cpp : implementation file
//
#include "stdafx.h"
#include "resource.h"
#include "MechThread.h"
#include "Audit.h"
#include "Trans.h"

#define AUDIT_FILE "audits.txt"
#define TRANS_LOG "trans.log"
#define TRANS_LOG_HDR "Serial #\tTransType\tDate\tDay\tTime\tCardSerial\tCnMax\tCnMin\tAmount\tRat1\tRat2\tClock\tErr\t\r\n"

#define MECH_DELAY 150
#define DEBUG_DELAY 10000

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CMechThread

#include "Comm.h"
#include "MeterCommDlg.h"

IMPLEMENT_DYNCREATE(CMechThread, CWinThread)

CMechThread::CMechThread()
{

	m_pAudit = NULL;
}

CMechThread::~CMechThread()
{

}

CMechThread::CMechThread(CComm *com)
{
	iLocationPtr = 300;

	iFunction = 0;
	bAlive = TRUE;
	m_bSuccess = FALSE;
	iState = 0;
	m_pComm = com;
	iStep = 0;
	m_pDlg = NULL;
	m_pTrans = NULL;
	m_pAudit = NULL;

}	


BOOL CMechThread::InitInstance()
{
	// TODO:  perform and per-thread initialization here

	
	return TRUE;
}

BOOL CMechThread::ReadRam(int iPage, int iByte, int iLen)
{
	
	CString csMsg;

	DWORD dwChecksum = 0;
	int i;

	if((iPage < 0) || (iPage > 255)) return(FALSE);
	if((iLen < 1) || (iLen > 256)) return(FALSE);
	if(iByte < 0) return(FALSE);

	Sleep(MECH_DELAY);

    m_pComm->WriteByte(1); // send command
	m_pComm->ReadBytes(1,MECH_TIMEOUT);

	if(m_pComm->cReadBuffer[0] != 1) {
		csMsg.Format(_T("RAM1 Command echo failed: %d / 1"),m_pComm->cReadBuffer[0]);
		Notify(&csMsg);
		return(FALSE); // if echo fails
	}

	m_pComm->WriteByte(iPage);
	m_pComm->ReadBytes(1,MECH_TIMEOUT);
	m_pComm->WriteByte(iByte);
	m_pComm->ReadBytes(1,MECH_TIMEOUT);
	m_pComm->WriteByte(iLen);
	m_pComm->ReadBytes(1,MECH_TIMEOUT);

	Sleep(MECH_DELAY);

	m_pComm->ReadBytes(1,MECH_TIMEOUT + 2); // read command

	Sleep(MECH_DELAY);

	if(m_pComm->cReadBuffer[0] != 1) {
		csMsg.Format(_T("Command echo failed: %d / 1"),m_pComm->cReadBuffer[0]);
		Notify(&csMsg);
		Sleep(DEBUG_DELAY);
		return(FALSE);
	}

	for(i=0;i<iLen;i++) {
		m_pComm->ReadBytes(1,MECH_TIMEOUT);
		csMsg.Format(_T("Reading %d/%d/%d"),iPage,iByte,iLen); // temporary
		Notify(&csMsg);

		if(m_pComm->sReadBufferLen < 1) {
			csMsg.Format(_T("Error reading mem (%d/%d/%d)"),iPage,iByte,iLen);
			Notify(&csMsg);
			Sleep(DEBUG_DELAY);
			return(FALSE);
		}
			
		cData[i] = m_pComm->cReadBuffer[0];
		dwChecksum = dwChecksum + cData[i];
	}

	m_pComm->ReadBytes(1,MECH_TIMEOUT); // read checksum

	if((dwChecksum & 255) != m_pComm->cReadBuffer[0]) {
		csMsg.Format(_T("CRC doesn't match (%d/%d/%d)"),iPage,iByte,iLen);
		Notify(&csMsg);
		Sleep(DEBUG_DELAY);
		return(FALSE);
	}

	m_pComm->ReadBytes(1,MECH_TIMEOUT); // get type char
	if(m_pComm->cReadBuffer[0] != 9) {
		csMsg.Format(_T("Final mt doesn't match: %d / 9"),m_pComm->cReadBuffer[0]);
		Notify(&csMsg);
		Sleep(DEBUG_DELAY);
		return(FALSE);
	}

	return(TRUE);
}


BOOL CMechThread::StartFunction(int iFunc)
{

	
	int iErr;
	CString csMsg;

	

	if(iErr = m_pComm->OpenPort(1)) {
		csMsg.Format(_T("Unable to open COM1: %d"),iErr);
		Notify(&csMsg);
		AfxMessageBox(csMsg);
		StopFunction();
		return(FALSE);
	}

	Sleep(400); // give the meter time to ack the RTS high
	
	iFunction = iFunc;

	switch(iFunction) {
		case FUNCTION_AUDIT:
			NotifySetRange(0,8);
			break;
		case FUNCTION_DL_TRANS:
			// who knows
			break;
		default:
			return(FALSE);
	}

	return(TRUE);
}

int CMechThread::OnIdle(LONG lCount)
{

	CString csMsg;
		
	switch(iFunction) {
	case 0:
			// we haven't gotten a function yet
		return(1);

	case FUNCTION_AUDIT:
		// disable security then run audit cmd
		switch(iState) {
		case STATE_SECURITY:
												ASSERT (TRUE == m_pComm->IsOpen());
		if(DisableSecurity(TRUE)) iState++;
				else NotifySetPos(0);
			break;
		case STATE_READAUDIT:
			if(ReadAudit(TRUE)) iState++;
			else {
				NotifySetPos(0);
				iState = 0;
			}
			break;
		case STATE_AUDITDONE:
			SaveAudit();
			StopFunction();
			NotifyStep();
			csMsg = "Done.";
			Notify(&csMsg);
			m_bSuccess = TRUE;
			return(0);
			break;
		}
	break;

	case FUNCTION_DL_TRANS:
		switch(iState) {
		case STATE_SECURITY:
			if(DisableSecurity(FALSE)) iState++;
				else NotifySetPos(0);
			break;

		case STATE_READAUDIT:
			if(ReadAudit(FALSE)) iState++;
			else {
				NotifySetPos(0);
				iState = 0;
			}
			break;
		case STATE_READTRANS:
			if(ReadTrans(TRUE)) iState++;
			else {
				Sleep(500); // time to read the display
				NotifySetPos(0);
				iState = 0;
			}
			break;

		case STATE_TRANSDONE:
			SaveTrans();
			StopFunction();
			csMsg = "Done. ";
			m_bSuccess = TRUE;
			return(0);

		}
		break;
	}

	return(1);
}


BOOL CMechThread::SaveAudit()
{


	if(!m_pAudit) return(FALSE);

	return(m_pAudit->Save());

}


BOOL CMechThread::SaveTrans()
{
	return(TRUE);
}

BOOL CMechThread::ReadTrans(BOOL b8KMech)
{
	/* we come into this function already security enabled and audited */

//	int iPageNum;
//	int iLocationPtr;

	unsigned char cTransHeader[16];
	int itMinutes;
	int itTensHours;

	int i;


	CString csMsg;
	CFile LogFile;

	char buf[255]; /* for writing to log file */

	CTrans *t;


	// first, read the stop address

	if(iLocationPtr == 300) {

	csMsg = "Reading stop address";
	Notify(&csMsg);

	if(!ReadRam(0,134,2)) return(FALSE);

	iStopPage = cData[1];
	iStopLocation = cData[0];

	// second, read the transaction page map

	csMsg = "Reading transaction page map";
	Notify(&csMsg);

	if(!ReadRam(128,122,1)) return(FALSE);

	iTransPage = cData[0];

	csMsg = "Reading transaction page map #2";
	Notify(&csMsg);

	if(iStopPage < iTransPage) {
		csMsg.Format(_T("Invalid Transaction Pointer"));
		return(FALSE);
	}

	

	if(!ReadRam(iTransPage,0,16)) return(FALSE);

#ifdef TRANS_DEBUG
	csMsg.Format(_T("Read from %d -> %d/%d"),iTransPage,iStopPage,iStopLocation);
	AfxMessageBox(csMsg);
#endif



	for(i=0;i<16;i++)
		cTransHeader[i] = cData[i];

	iTransNum = cData[9] + (cData[10] * 256) + (cData[11] * 65536); /* last bit may be commented out ? */

	iEndPage = iStopPage;
	iEndLocation = iStopLocation;

	if(cData[0] != 0) {
		bOverflow = TRUE;

		// we have hardcoded a 8K mech here

		// later maybe I'll 
		if(b8KMech) {
			iStopPage = 153;
			iStopLocation = 0;
			iAddTransCntr = ((iStopPage - iTransPage) * 16) - 1;
		} else {
			iStopPage = 246;
			iStopLocation = 0;
			iAddTransCntr = ((iStopPage - iTransPage) * 16) - 1;
		}
	} else {
		bOverflow = FALSE;
	}

	/* I quote from the original source:

					'if rolled over, read entire database area by
                    'reading thru page 240
                    'check for 246 when using 32k E^2
                    'set this source code to 153 when using 8k E^2

    
	*/

	
	} /* end of if we don't know it all */

	LogFile.Open(_T(TRANS_LOG),CFile::modeCreate | CFile::modeWrite | CFile::modeNoTruncate,NULL);

	LogFile.Write(TRANS_LOG_HDR,strlen(TRANS_LOG_HDR));
	
	iPageNum = iTransPage;
	iLocationPtr = 16;

	/* Start reading transaction pages */


	NotifySetRange((16 * iPageNum),(16 * iStopPage));

	csMsg.Format(_T("Entering while((%d != %d) && (%d < %d))"),iPageNum,iStopPage,iLocationPtr,iStopLocation);
	
	while(!((iPageNum == iStopPage) && (iLocationPtr >= iStopLocation))) {
		csMsg.Format(_T("Reading.. (%d/%d)"),iPageNum,iLocationPtr);
		Notify(&csMsg);

		NotifySetPos(((16 * iPageNum) + iLocationPtr));

	if(!ReadRam(iPageNum, iLocationPtr,16)) {
		sprintf(buf,"ReadTrans Transrecord didn't read RAM @ %d, %d",iPageNum, iLocationPtr);
		LogFile.Write(buf,strlen(buf));
		return(FALSE);
	}

	iLocationPtr = iLocationPtr + 16;
	if (iLocationPtr == 256) {
		iLocationPtr = 0;
		iPageNum++;
	}

	if((cData[13] == 255) && (cData[14] == 255)) {
		/* null record */
		csMsg.Format(_T("Skipping null message: %d/%d"),iPageNum,iLocationPtr);
		Notify(&csMsg);

	} else {
		/* I quote from the original source:
			'Increment the transaction counter (increment at the beginning
			'   because number stored in meter is count of last record) */

		iTransNum++;
		if(bOverflow) {
			if(iPageNum < iEndPage) {
				ilTransNum = iTransNum + iAddTransCntr;
			} else {
				if((iPageNum == iEndPage) && (iLocationPtr < iEndLocation)) {
					ilTransNum = iTransNum + iAddTransCntr;
					iLastTransNum = ilTransNum;
				} else {
					ilTransNum = iTransNum;
				}
			}
		} else {
			ilTransNum = iTransNum;
		}

		csMsg.Format(_T("Reading transaction: %d"),ilTransNum);
		Notify(&csMsg);

		sprintf(buf,"%6d\t",m_pAudit->iSerialNumber);
		LogFile.Write(buf,strlen(buf));
		sprintf(buf,"%11d\t",ilTransNum);
		LogFile.Write(buf,strlen(buf));
		
		if(!m_pTrans) {
			m_pTrans = new CTrans(NULL);
			t = m_pTrans;
		} else {
			t = new CTrans(m_pTrans);
		}

		/* at this point we're committed, so let's make our object */
		
		t->iSerialNumber = m_pAudit->iSerialNumber;
		t->iTransNumber = ilTransNum;

		t->iType = cData[0];
		sprintf(buf,"%d\t",t->iType);
		LogFile.Write(buf,strlen(buf));

		t->iDay = (cData[6] / 32) + 1;
		sprintf(buf,"%d\t",t->iDay);
		LogFile.Write(buf,strlen(buf));


		/* Again I must quote from the source code: 
			'transaction time
			'hours = (giData(6) And 31)
			'minute = Int(giData(5) / 16) * 10 + (giData(5) And 15)
			'seconds = Int(giData(4) / 2)
		*/

		sprintf(t->cTime,"%2d:%2d:%2d",(cData[6] & 31),(((cData[5] /16) * 10) + (cData[5] & 15)),(cData[4]/2));
		sprintf(buf,"%s\t",t->cTime);
		LogFile.Write(buf,strlen(buf));

		/* serial number of card */

		if((t->iType == 10) || (t->iType == 11)) {
			t->dwCoinMax = cData[7] + (cData[8] * 256);
			t->dwCoinMin = cData[9] + (cData[10] * 256);
		} else {
			t->dwCardSN = cData[7] + (cData[8] * 256) + (cData[9] * 65536) + ((cData[10] & 63) * 1677216);
		}

		sprintf(buf,"%10d\t",t->dwCardSN);
		LogFile.Write(buf,strlen(buf));
		sprintf(buf,"%5d\t",t->dwCoinMax);
		LogFile.Write(buf,strlen(buf));
		sprintf(buf,"%5d\t",t->dwCoinMin);

		switch(t->iType) {
		case 11:
			t->PerDevError = 0;
			t->iAmount = 0;
			t->CoinRat1 = cData[11];
			t->CoinRat2 = cData[12];
			break;
		case 16:
			t->PerDevError = cData[12];
			break;
		default:
			t->iAmount = (cData[11] + (cData[12] * 256));
		}

		sprintf(buf,"%10d\t",t->iAmount);
		LogFile.Write(buf,strlen(buf));
		sprintf(buf,"%3d\t",t->CoinRat1);
		LogFile.Write(buf,strlen(buf));
		sprintf(buf,"%3d\t",t->CoinRat2);
		LogFile.Write(buf,strlen(buf));
		sprintf(buf,"%3d\t",t->PerDevError);
		LogFile.Write(buf,strlen(buf));

		/* MOre from the original source:
'transaction parking clock
        'If the ms nibble of the parking clock hours digit (14) is
        '< 10, then the parking clock was in paid time and the ms
        'nibble represents the tens of hours on the parking clock
        'If the ms nibble is ten or more, the parking clock was in
        'overtime and the first parking clock digit was "-" and the
        'ls nibble indicates the hours digit of overtime */

	
		itMinutes = (((cData[13] / 16) & 15) * 10) + (cData[13] & 15);
		itTensHours = ((cData[14] / 16) & 15);
		if(itTensHours < 10) {
			t->iParkingClock = itMinutes + ((cData[14] & 15)  * 100) + (itTensHours * 1000);
		} else {
			t->iParkingClock = -1 * (((cData[14] & 15) * 100) + itMinutes);
		}

		sprintf(buf,"%4d\t",t->iParkingClock);
		LogFile.Write(buf,strlen(buf));

		t->iError = cData[15];
		sprintf(buf,"%2d\r\n",t->iError);
		LogFile.Write(buf,strlen(buf));

	} /* if null record */

	}	/* master while */

	csMsg.Format(_T("Exiting read while() loop"));
	Notify(&csMsg);

	LogFile.Close();

	return(TRUE); /* whew! */

}

BOOL CMechThread::StopFunction()
{
	m_pDlg = NULL; /* stop callbacks */
	bAlive = FALSE;
	m_pComm->ClosePort();
	PostQuitMessage(0);

	return(TRUE);
}

BOOL CMechThread::Notify(CString *text)
{
	return(NotifyText(text));
}

BOOL CMechThread::ReadAudit(BOOL bStep)
{
	CString csMsg;
	
	if(bStep) NotifyStep();
	Sleep(MECH_DELAY); // give the mech a chance to stablize
	csMsg = "Sending audit cmd";
	Notify(&csMsg);

	m_pComm->WriteByte(3);
	
	if(bStep) NotifyStep();
	csMsg = "Reading audit reply";
	Notify(&csMsg);
	
	m_pComm->ReadBytes(20,MECH_TIMEOUT);
	
	if(m_pComm->sReadBufferLen < 18) {
		csMsg = "Failed to read audit";
		Notify(&csMsg);
		return(FALSE);
	}

	if(m_pAudit) delete m_pAudit;

	m_pAudit = new CAudit;

#ifdef AUDIT_DEBUG
	csMsg.Format(_T("%d %d %d %d %d %d %d %d\n%d %d %d %d %d %d %d %d"),m_pComm->cReadBuffer[0],m_pComm->cReadBuffer[1]
											 ,m_pComm->cReadBuffer[2],m_pComm->cReadBuffer[3]
											 ,m_pComm->cReadBuffer[4],m_pComm->cReadBuffer[5]
											 ,m_pComm->cReadBuffer[6],m_pComm->cReadBuffer[7]
											 ,m_pComm->cReadBuffer[8],m_pComm->cReadBuffer[9],
											 m_pComm->cReadBuffer[10],m_pComm->cReadBuffer[11],
											 m_pComm->cReadBuffer[12],m_pComm->cReadBuffer[13],
											 m_pComm->cReadBuffer[14],m_pComm->cReadBuffer[15]);
	AfxMessageBox(csMsg);
#endif

	if(m_pComm->cReadBuffer[0] != 3) return(FALSE);
	
#define CORR 1

	m_pAudit->iSerialNumber = m_pComm->cReadBuffer[2] + 
		(m_pComm->cReadBuffer[3] * 100) + (m_pComm->cReadBuffer[4] * 10000);

	m_pAudit->csSerialNumber.Format(_T("%6d"),m_pAudit->iSerialNumber);
	m_pAudit->iNycLocation = m_pComm->cReadBuffer[(CORR + 5)];
	m_pAudit->iCityCode = m_pComm->cReadBuffer[(CORR + 8)];
	m_pAudit->iScale = m_pComm->cReadBuffer[(CORR + 10)];
	m_pAudit->iAuditRegister = CalcValue(m_pComm->cReadBuffer+(CORR + 11),2);
	m_pAudit->iAuditAmount = m_pAudit->iAuditRegister * m_pAudit->iScale;
	m_pAudit->iLogError = m_pComm->cReadBuffer[(CORR+15)];
	m_pAudit->iCurrentError = m_pComm->cReadBuffer[(CORR+16)];
	
	m_pAudit->d = div(m_pComm->cReadBuffer[(CORR+14)],20);


	m_pAudit->csBatteryLevel.Format(_T("%d.%d"),m_pAudit->d.quot,m_pAudit->d.rem);

	 if(bStep) NotifyStep();

	return(TRUE);

}

int CMechThread::CalcValue(unsigned char *c_pPtr,int iLen) 
{

int i;
int Value = 0;
int Multiplier;

	for(i=0,Multiplier=1; i<iLen; i++,Multiplier*=256) {
		Value+=(unsigned int)(c_pPtr[i]) * (unsigned int)(Multiplier);
	}

	return Value;
}


BOOL CMechThread::DisableSecurity(BOOL bStep)
{
	// attempt to disable security on the mech
	// 4 steps

	CString csMsg;
	const int revbit[16] = {7,2,11,10,13,8,0,14,12,6,4,1,15,3,5,9};
	short aEncrypt[9];
	int i;

	csMsg = "Enabling communications";

	Notify(&csMsg);

	Sleep(MECH_DELAY); // give the meter time to clear

	m_pComm->WriteByte(6);

	if (bStep) NotifyStep();

	m_pComm->ReadBytes(7,MECH_TIMEOUT);

	Sleep(75);

	if (bStep) NotifyStep();

	if(m_pComm->sReadBufferLen < 7) {
		csMsg = "No meter found!";
		Notify(&csMsg);
		return(FALSE);
	}
	
	for(i=0;i<4;i++) {
				aEncrypt[2*i] = m_pComm->cReadBuffer[i] & 15;
				aEncrypt[(2*i)+1] = m_pComm->cReadBuffer[i] / 16;
	}

	aEncrypt[0] = revbit[(aEncrypt[0] ^ aEncrypt[4])];
	aEncrypt[1] = revbit[(aEncrypt[1] ^ aEncrypt[7])];
	aEncrypt[2] = revbit[(aEncrypt[2] ^ aEncrypt[6])];
	aEncrypt[3] = revbit[(aEncrypt[3] ^ aEncrypt[5])];

	if(bStep) NotifyStep();

	m_pComm->WriteByte(7);

	m_pComm->ReadBytes(1,MECH_TIMEOUT);
	m_pComm->WriteByte(aEncrypt[1]);
	m_pComm->ReadBytes(1,MECH_TIMEOUT);
	m_pComm->WriteByte(aEncrypt[3]);
	m_pComm->ReadBytes(1,MECH_TIMEOUT);
	m_pComm->WriteByte(aEncrypt[0]);
	m_pComm->ReadBytes(1,MECH_TIMEOUT);
	m_pComm->WriteByte(aEncrypt[2]);
	m_pComm->ReadBytes(1,MECH_TIMEOUT);
	m_pComm->ReadBytes(1,MECH_TIMEOUT); // get devtype

	if(bStep) NotifyStep();


	if((m_pComm->cReadBuffer[0] != 9) || m_pComm->sReadBufferLen < 1) {
		csMsg = "Failed enabling security";
		Notify(&csMsg);
		return(FALSE);
	}

	return(TRUE);

}

BOOL CMechThread::NotifyStep() {
	iStep++;
	return(NotifySetPos(iStep));
}

BOOL CMechThread::NotifySetRange(int iLow, int iHigh) 
{
	if(!m_pDlg) return FALSE;

	return(m_pDlg->NotifySetRange(iLow,iHigh));
}

BOOL CMechThread::NotifyText(CString *text)
{
	return(m_pDlg->NotifyText(text));
}

BOOL CMechThread::NotifySetPos(int iPos)
{
	iStep = iPos;

	return(m_pDlg->NotifySetPos(iPos));
}

int CMechThread::ExitInstance()
{
	// TODO:  perform any per-thread cleanup here
	

	return CWinThread::ExitInstance();
}

BEGIN_MESSAGE_MAP(CMechThread, CWinThread)
	//{{AFX_MSG_MAP(CMechThread)
		// NOTE - the ClassWizard will add and remove mapping macros here.
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CMechThread message handlers
